from __future__ import print_function, unicode_literals

import re
from getpass import getuser

from algen.consts import POSTGRES_TYPES, MUTABLE_DICT_TYPES, NO_PARAMS_TYPES
from algen.templates import ALCHEMY_TEMPLATES
from .logger import LOG

__author__ = "danishabdullah"

TYPE_INFO_REGEX = re.compile("(^\w+)(\(.*\)$)?")

__all__ = ('ModelCompiler',)


class ModelCompiler(object):
    """
    Class for compiling and producing a Sql Alchemy model from user provided definition.

    Everything is implemented as a property for rapid development without any  thought
    for speed.
    """

    def __init__(self, name, column_def):
        assert isinstance(column_def, dict)
        self.model_def = {'name': column_def}
        self.username = getuser()
        self.tab = '    '
        self.class_name = name
        self.column_definitions = column_def['columns']
        self.foreign_key_definitions = column_def.get('foreign_keys', [])
        self.relationship_definitions = column_def.get('relationships', [])

    @staticmethod
    def convert_case(name):
        """Converts name from CamelCase to snake_case"""
        LOG.info("Converting name from CamelCase to snake_case.")
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    @property
    def table_name(self):
        """Pluralises the class_name using utterly simple algo and returns as table_name"""
        LOG.info("Generating table name.")
        if not self.class_name:
            raise ValueError
        else:
            tbl_name = ModelCompiler.convert_case(self.class_name)
        last_letter = tbl_name[-1]
        if last_letter in ("y",):
            return "{}ies".format(tbl_name[:-1])
        elif last_letter in ("s",):
            return "{}es".format(tbl_name)
        else:
            return "{}s".format(tbl_name)

    @staticmethod
    def get_column_type(string):
        """Extracts and compiles column info"""
        LOG.info("Extracting column type info.")
        return ModelCompiler.get_col_type_info(string)[0]

    @staticmethod
    def get_type_params(string):
        """Extracts type params from column type definition"""
        LOG.info("Extracting column type params info.")
        return ModelCompiler.get_col_type_info(string)[1]

    @staticmethod
    def get_col_type_info(string):
        """Extracts column type information from column definition string"""
        LOG.info("Extracting type and params info from the spec")
        for match in TYPE_INFO_REGEX.finditer(string):
            name, params = match.groups()
            return name if name else '', params if params else ''
        return '', ''

    @property
    def types(self):
        """All the unique types found in user supplied model"""
        LOG.info("Extracting all the types referenced in the spec.")
        res = []
        for column in self.column_definitions:
            tmp = column.get('type', None)
            res.append(ModelCompiler.get_column_type(tmp)) if tmp else False
        res = list(set(res))
        return res

    @property
    def postgres_types(self):
        """Returns known postgres only types referenced in user supplied model"""
        LOG.info("Extracting all the postgres specific types referenced in the spec.")
        return [n for n in self.types if n in POSTGRES_TYPES]

    @property
    def standard_types(self):
        """Returns non-postgres types referenced in user supplied model"""
        LOG.info("Extracting all the standard types referenced in the spec.")
        return [n for n in self.types if n not in POSTGRES_TYPES]

    @property
    def basic_types(self):
        """Returns non-postgres types referenced in user supplied model """
        LOG.info("Extracting all the basic types referenced in the spec.")
        if not self.foreign_key_definitions:
            return self.standard_types
        else:
            tmp = self.standard_types
            tmp.append('ForeignKey')
            return tmp

    @property
    def mutable_dict_types(self):
        """Return only mutable dict types referenced in the spec"""
        LOG.info("Extracting all the types that can be modelled as mutable_dicts.")
        return [n for n in self.types if n in MUTABLE_DICT_TYPES]

    @property
    def primary_keys(self):
        """Returns the primary keys referenced in user supplied model"""
        LOG.info("Extracting all the primary keys referenced in the spec.")
        res = []
        for column in self.column_definitions:
            if 'primary_key' in column.keys():
                tmp = column.get('primary_key', None)
                res.append(column['name']) if tmp else False
        return res

    @property
    def compiled_named_imports(self):
        """Returns compiled named imports required for the model"""
        LOG.info("Compiling all the named imports.")
        res = []
        if self.postgres_types:
            res.append(
                ALCHEMY_TEMPLATES.named_import.safe_substitute(
                    module='sqlalchemy.dialects.postgresql',
                    labels=", ".join(self.postgres_types)))
        if self.mutable_dict_types:
            res.append(
                ALCHEMY_TEMPLATES.named_import.safe_substitute(
                    module='sqlalchemy.ext.mutable', labels='MutableDict'
                ))
        return "\n".join(res)

    @property
    def compiled_orm_imports(self):
        """Returns compiled named imports required for the model"""
        LOG.info("Compiling all the orm imports.")
        module = 'sqlalchemy.orm'
        labels = []
        if self.relationship_definitions:
            labels.append("relationship")
        return ALCHEMY_TEMPLATES.named_import.safe_substitute(module=module, labels=", ".join(labels))

    @property
    def compiled_columns(self):
        """Returns compiled column definitions"""

        def get_column_args(column):
            tmp = []
            for arg_name, arg_val in column.items():
                if arg_name not in ('name', 'type'):
                    if arg_name in ('server_default', 'server_onupdate'):
                        arg_val = '"{}"'.format(arg_val)
                    tmp.append(ALCHEMY_TEMPLATES.column_arg.safe_substitute(arg_name=arg_name,
                                                                            arg_val=arg_val))
            return ", ".join(tmp)

        LOG.info("Compiling all the column definitions.")
        res = []
        for column in self.column_definitions:
            column_args = get_column_args(column)
            column_type, type_params = ModelCompiler.get_col_type_info(column.get('type'))
            column_name = column.get('name')
            if column_type in MUTABLE_DICT_TYPES:
                column_type = ALCHEMY_TEMPLATES.mutable_dict_type.safe_substitute(type=column_type,
                                                                                  type_params=type_params)
                type_params = ''
            res.append(
                ALCHEMY_TEMPLATES.column_definition.safe_substitute(column_name=column_name,
                                                                    column_type=column_type,
                                                                    column_args=column_args,
                                                                    type_params=type_params))
        join_string = "\n" + self.tab
        return join_string.join(res)

    @property
    def compiled_foreign_keys(self):
        """Returns compiled foreign key definitions"""

        def get_column_args(column):
            tmp = []
            for arg_name, arg_val in column.items():
                if arg_name not in ('name', 'type', 'reference'):
                    if arg_name in ('server_default', 'server_onupdate'):
                        arg_val = '"{}"'.format(arg_val)
                    tmp.append(ALCHEMY_TEMPLATES.column_arg.safe_substitute(arg_name=arg_name,
                                                                            arg_val=arg_val))
            return ", ".join(tmp)

        def get_fkey_args(column):
            table = column['reference']['table']
            column = column['reference']['column']
            return ALCHEMY_TEMPLATES.foreign_key_arg.safe_substitute(reference_table=table, reference_column=column)

        LOG.info("Compiling all the foreign key definitions.")
        res = []
        for column in self.foreign_key_definitions:
            column_args = get_column_args(column)
            column_type, type_params = ModelCompiler.get_col_type_info(column.get('type'))
            column_name = column.get('name')
            reference = get_fkey_args(column)
            if column_type in MUTABLE_DICT_TYPES:
                column_type = ALCHEMY_TEMPLATES.mutable_dict_type.safe_substitute(type=column_type,
                                                                                  type_params=type_params)
                type_params = ''
            res.append(
                ALCHEMY_TEMPLATES.foreign_key.safe_substitute(column_name=column_name,
                                                              column_type=column_type,
                                                              column_args=column_args,
                                                              foreign_key_args=reference,
                                                              type_params=type_params))
        join_string = "\n" + self.tab
        return join_string.join(res)

    @property
    def compiled_relationships(self):
        """Returns compiled relationship definitions"""

        def get_column_args(column):
            tmp = []
            for arg_name, arg_val in column.items():
                if arg_name not in ('name', 'type', 'reference', 'class'):
                    if arg_name in ('back_populates', ):
                        arg_val = "'{}'".format(arg_val)
                    tmp.append(ALCHEMY_TEMPLATES.column_arg.safe_substitute(arg_name=arg_name,
                                                                            arg_val=arg_val))
            return ", ".join(tmp)

        LOG.info("Compiling all the relationship definitions.")
        res = []
        for column in self.relationship_definitions:
            column_args = get_column_args(column)
            column_name = column.get('name')
            cls_name = column.get("class")
            res.append(
                ALCHEMY_TEMPLATES.relationship.safe_substitute(column_name=column_name,
                                                               column_args=column_args,
                                                               class_name=cls_name))
        join_string = "\n" + self.tab
        return join_string.join(res)

    @property
    def columns(self):
        """Return names of all the addressable columns (including foreign keys) referenced in user supplied model"""
        LOG.info("Extracting all the database column names from the spec.")
        res = [col['name'] for col in self.column_definitions]
        res.extend([col['name'] for col in self.foreign_key_definitions])
        return res

    @property
    def compiled_init_func(self):
        """Returns compiled init function"""

        def get_column_assignment(column_name):
            return ALCHEMY_TEMPLATES.col_assignment.safe_substitute(col_name=column_name)

        def get_compiled_args(arg_name):
            return ALCHEMY_TEMPLATES.func_arg.safe_substitute(arg_name=arg_name)

        LOG.info("Compiling the __init__ function.")
        join_string = "\n" + self.tab + self.tab
        column_assignments = join_string.join([get_column_assignment(n) for n in self.columns])
        init_args = ", ".join(get_compiled_args(n) for n in self.columns)
        return ALCHEMY_TEMPLATES.init_function.safe_substitute(col_assignments=column_assignments,
                                                               init_args=init_args)

    @property
    def compiled_update_func(self):
        """Returns compiled update function"""

        def get_not_none_col_assignment(column_name):
            return ALCHEMY_TEMPLATES.not_none_col_assignment.safe_substitute(col_name=column_name)

        def get_compiled_args(arg_name):
            return ALCHEMY_TEMPLATES.func_arg.safe_substitute(arg_name=arg_name)

        LOG.info("Compiling the update function.")
        join_string = "\n" + self.tab + self.tab
        columns = [n for n in self.columns if n not in self.primary_keys]
        not_none_col_assignments = join_string.join([get_not_none_col_assignment(n) for n in columns])
        update_args = ", ".join(get_compiled_args(n) for n in columns)
        return ALCHEMY_TEMPLATES.update_function.safe_substitute(not_none_col_assignments=not_none_col_assignments,
                                                                 update_args=update_args,
                                                                 class_name=self.class_name)

    @property
    def compiled_hash_func(self):
        """Returns compiled hash function based on hash of stringified primary_keys.
        This isn't the most efficient way"""

        def get_primary_key_str(pkey_name):
            return "str(self.{})".format(pkey_name)

        LOG.info("Compiling the __hash__ function.")
        hash_str = "+ ".join([get_primary_key_str(n) for n in self.primary_keys])
        return ALCHEMY_TEMPLATES.hash_function.safe_substitute(concated_primary_key_strs=hash_str)

    def comparator_compiler(self, negation_type):
        def get_primary_key_comparator(pkey_name):
            return ALCHEMY_TEMPLATES.key_col_comparator.safe_substitute(col=pkey_name)

        LOG.info("Compiling a comparator function.")
        assert negation_type in ('positive', 'negative')
        negation_marker = "not " if negation_type == 'negative' else ""
        func_name = "__neq__" if negation_type == 'negative' else "__eq__"
        key_comparators = [get_primary_key_comparator(n) for n in self.primary_keys]
        key_comparators = " and ".join(key_comparators)
        return ALCHEMY_TEMPLATES.comparator_function.safe_substitute(func_name=func_name,
                                                                     negation_marker=negation_marker,
                                                                     key_col_comparisons=key_comparators)

    @property
    def compiled_eq_func(self):
        """Returns compiled equality function"""
        LOG.info("Compiling the __eq__ function.")
        return self.comparator_compiler('positive')

    @property
    def compiled_neq_func(self):
        LOG.info("Compiling the __neq__ function.")
        return self.comparator_compiler('negative')

    def representation_function_compiler(self, func_name):
        """Generic function can be used to compile __repr__ or __unicode__ or __str__"""

        def get_col_accessor(col):
            return ALCHEMY_TEMPLATES.col_accessor.safe_substitute(col=col)

        def get_col_evaluator(col):
            return ALCHEMY_TEMPLATES.col_evaluator.safe_substitute(col=col)

        LOG.info("Compiling a representation function.")
        col_evaluators = ", ".join([get_col_evaluator(n) for n in self.primary_keys])
        col_accessors = ", ".join([get_col_accessor(n) for n in self.primary_keys])

        return ALCHEMY_TEMPLATES.representor_function.safe_substitute(func_name=func_name,
                                                                      col_accessors=col_accessors,
                                                                      col_evaluators=col_evaluators,
                                                                      class_name=self.class_name)

    @property
    def compiled_str_func(self):
        """Returns compiled __str__ function"""
        LOG.info("Compiling the __str__ function.")
        return self.representation_function_compiler('str')

    @property
    def compiled_unicode_func(self):
        """Return compiled __unicode__ function"""
        LOG.info("Compiling the __unicode__ function.")
        return self.representation_function_compiler('unicode')

    @property
    def compiled_repr_func(self):
        """Returns compiled __repr__ function"""
        LOG.info("Compiling the __repr__ function.")
        return self.representation_function_compiler('repr')

    @property
    def compiled_proxy_cls_func(self):
        """Returns compile get_proxy_cls function"""
        LOG.info("Compiling the get_proxy_cls function.")
        return ALCHEMY_TEMPLATES.get_proxy_cls_function.safe_substitute(class_name=self.class_name)

    @property
    def compiled_model(self):
        """Returns compile ORM class for the user supplied model"""
        LOG.info("Compiling the < > Model.".format(self.class_name, self.table_name))
        return ALCHEMY_TEMPLATES.model.safe_substitute(class_name=self.class_name,
                                                       table_name=self.table_name,
                                                       column_definitions=self.compiled_columns,
                                                       init_function=self.compiled_init_func,
                                                       update_function=self.compiled_update_func,
                                                       hash_function=self.compiled_hash_func,
                                                       eq_function=self.compiled_eq_func,
                                                       neq_function=self.compiled_neq_func,
                                                       str_function=self.compiled_str_func,
                                                       unicode_function=self.compiled_unicode_func,
                                                       repr_function=self.compiled_repr_func,
                                                       types=", ".join(self.basic_types),
                                                       username=self.username,
                                                       foreign_keys=self.compiled_foreign_keys,
                                                       relationships=self.compiled_relationships,
                                                       named_imports=self.compiled_named_imports,
                                                       orm_imports=self.compiled_orm_imports,
                                                       get_proxy_cls_function=self.compiled_proxy_cls_func,
                                                       add_function=ALCHEMY_TEMPLATES.add_function.template,
                                                       delete_function=ALCHEMY_TEMPLATES.delete_function.template,
                                                       to_dict_function=ALCHEMY_TEMPLATES.to_dict_function.template,
                                                       to_proxy_function=ALCHEMY_TEMPLATES.to_proxy_function.template,
                                                       from_proxy_function=ALCHEMY_TEMPLATES.from_proxy_function.template)
